#!/usr/bin/python3
import sys
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.font
import pygubu.widgets.fontinputui as baseui

from contextlib import suppress
from pygubu.utils.widget import WidgetConfigureMixin
from pygubu.utils.font import tkfontstr_to_dict
from pygubu.component.style_manager import IStyleDefinition


PREDEFINED_FONTS = [
    "TkDefaultFont",
    "TkTextFont",
    "TkFixedFont",
    "TkMenuFont",
    "TkHeadingFont",
    "TkCaptionFont",
    "TkSmallCaptionFont",
    "TkIconFont",
    "TkTooltipFont",
]

_sp = sys.platform
if _sp in ("win32", "cygwin"):
    PREDEFINED_FONTS.extend(
        ("system", "ansi", "device", "systemfixed", "ansifixed", "oemfixed")
    )
if _sp == "darwin":
    PREDEFINED_FONTS.extend(
        (
            "system",
            "application",
            "menu",
            "systemSystemFont",
            "systemEmphasizedSystemFont",
            "systemSmallSystemFont",
            "systemSmallEmphasizedSystemFont",
            "systemApplicationFont",
            "systemLabelFont",
            "systemViewsFont",
            "systemMenuTitleFont",
            "systemMenuItemFont",
            "systemMenuItemMarkFont",
            "systemMenuItemCmdKeyFont",
            "systemWindowTitleFont",
            "systemPushButtonFont",
            "systemUtilityWindowTitleFont",
            "systemAlertHeaderFont",
            "systemToolbarFont",
            "systemMiniSystemFont",
            "systemDetailSystemFont",
            "systemDetailEmphasizedSystemFont",
        )
    )


class Icon:
    WEIGHT: str = "weight"
    SLANT: str = "slant"
    UNDERLINE: str = "underline"
    OVERSTRIKE: str = "overstrike"
    _KDATA_PNG = 1
    _KDATA_GIF = 2
    _KIMG = 3
    _META = {
        WEIGHT: {
            _KDATA_PNG: b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAA3XAAAN1wFCKJt4AAAApElEQVQ4y83SMQ5BQRSF4Q8lhVIUWpFoLMACWAClvehtwibUCo1EdKLQSHSUJF5EnmYkIooxDSeZTM7M5L9zTy6/VuHDWRe1F3/HFptY6A35hzVFMQaQY/DiS+jhitH74xjiHTPMQ3tfA6CKNvYxIeZYYxd8B3VkaOIY84PsrYUDypikhPhUP9y1UjKARdgbKYAKxrhglTpIZwxTRlmovMTJ3+kBfHQpHDGWZagAAAAASUVORK5CYII=",
            _KDATA_GIF: b"R0lGODlhEAAQAIABAAAAAP///yH5BAEKAAEALAAAAAAQABAAAAIfjI+py70AI2gT1Zcgu4ZnqW1gqHiB2ZXbJ4LOC8dLAQA7",
        },
        SLANT: {
            _KDATA_PNG: b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAA3XAAAN1wFCKJt4AAAAj0lEQVQ4y83SMQ4BURCH8Z8tcASNiMYdFCoaLsLxREGlcgeFZqPTU5AtaKYQid31FEwymeJNvvle8ufX1Sh5a2MWs48DTtjVhQ9R4P7UeYrlCsvULzZxxjwVMAn13ruFrAIwxR7HbwCbVP1u6I9TAQtc0Cpbyir0t7h9ejl/CU8Roaod5RE6GATsinXMP6wH1K8dDvLj3yoAAAAASUVORK5CYII=",
            _KDATA_GIF: b"R0lGODlhEAAQAIABAAAAAP///yH5BAEKAAEALAAAAAAQABAAAAIajI+py+0AoTNyUnsADtX2qW2hNzbRt6WqUgAAOw==",
        },
        UNDERLINE: {
            _KDATA_PNG: b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAA3XAAAN1wFCKJt4AAAAl0lEQVQ4y83TMQrCQBCF4S9RIZfRUrGyt5JcRy+lgq2Nd9B72AhBtJkikY0JCJIHU+z/lsfs7ixD1BJFghfhdapCmeBleA3liY3jqF48//W8qYBnCx+F1xlwwzTBZ7j26WqHO+Y1tgi27RMwwTFu/BJV4RBeQ1lLSIYNVrE+Y/+XCazw6qiqPpHZx6iu47m+6YkTHsP4eW9nXx7XH08P8gAAAABJRU5ErkJggg==",
            _KDATA_GIF: b"R0lGODlhEAAQAIABAAAAAP///yH5BAEKAAEALAAAAAAQABAAAAIdjI+py72g4ItMIpuw0Qdg33yOR3KR6YTlmrbuGxQAOw==",
        },
        OVERSTRIKE: {
            _KDATA_PNG: b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAA3XAAAN1wFCKJt4AAAAyUlEQVQ4y83Sv0rCYRTG8Y8aKC3SHiSCQ0GBY0OjEXQPgpfR4Cx4DQ1BV9AQBN1GhEsRPxC3oKZ+g3+WM8QPlVcXfeCFF855vzznOS+7VmVN7QxXaGGK71RoGY+YF84TaimAHv5wEw6ruEaGyxTAPV42sVtUhnPUUwAH/+4NDHCII0wiuBl+8YE8Ar3D1yoHOV7xGY/LaKITeWylOsbop2SwTD94w0lK8xDPOI01VnAbo3VTABcxf/EjPSxzXFqznTaOo+cdI3upBXtuKCA7o2yvAAAAAElFTkSuQmCC",
            _KDATA_GIF: b"R0lGODlhEAAQAIABAAAAAP///yH5BAEKAAEALAAAAAAQABAAAAIdjI+py80AowOIPusOnrFD5XmZsWXfcpJcN7auWwAAOw==",
        },
    }

    @classmethod
    def load(cls, icon_id, master=None):
        icon = None
        if icon_id in cls._META:
            icon = cls._META[icon_id].get(cls._KIMG, None)
            if icon is None:
                data_key = cls._KDATA_GIF
                if tk.TkVersion >= 8.6:
                    data_key = cls._KDATA_PNG
                icon = tk.PhotoImage(
                    data=cls._META[icon_id][data_key], master=master
                )
                cls._META[icon_id][cls._KIMG] = icon
        return icon


class FontInputStyleManager(IStyleDefinition):
    UID = "FontInput"

    def setup(self, style: ttk.Style) -> None:
        buttons = (Icon.WEIGHT, Icon.SLANT, Icon.UNDERLINE, Icon.OVERSTRIKE)
        for bname in buttons:
            style_name = f"{bname}.{self.UID}.Toolbutton"
            icon = Icon.load(bname)
            style.configure(style_name, image=icon)


class FontInput(WidgetConfigureMixin, baseui.FontInputUI):
    EVENT_FONT_CHANGED = "<<FontInput:FontChanged>>"
    KEY_PRESS_MS = 850
    FONT_SIZE_DEFAULT = None
    SM = FontInputStyleManager()

    def __init__(self, master=None, variable: tk.Variable = None, **kw):
        super().__init__(master, **kw)
        tkvar = tk.StringVar(self, "") if variable is None else variable
        self._font_var = tkvar  # The current font value
        self._kp_cb = None
        self._fp_cb = None
        self._init_font_size()
        self._populate_options()
        self.SM.initialize(self)

    @classmethod
    def _init_font_size(cls):
        if cls.FONT_SIZE_DEFAULT is None:
            font: tk.font.Font = tk.font.nametofont("TkDefaultFont")
            size = font.cget("size")
            cls.FONT_SIZE_DEFAULT = abs(size)
            # print(abs(size), font.actual())

    def _widget_cget(self, option):
        if option == "value":
            return self._font_var.get()
        if option == "variable":
            return self._entry_var
        return super()._widget_cget(option)

    def _configure_get(self, option):
        if option in ("value", "variable"):
            return self._widget_cget(option)
        return super()._configure_get(option)

    def _configure_set(self, **kw):
        key = "variable"
        if key in kw:
            self._font_var = kw.pop(key)
            self._convert_to_form()
        key = "value"
        if key in kw:
            value = kw.pop(key)
            self._font_var.set(value)
            self._convert_to_form()
        key = "state"
        if key in kw:
            new_state = kw.pop(key)
            if new_state in ("normal", "disabled"):
                self.wfamily.configure(state=new_state)
                self.wsize.configure(state=new_state)
                self.wweight.configure(state=new_state)
                self.wslant.configure(state=new_state)
                self.wunderline.configure(state=new_state)
                self.woverstrike.configure(state=new_state)
        return super()._configure_set(**kw)

    def _populate_options(self):
        sizes = (
            5,
            6,
            8,
            9,
            10,
            11,
            12,
            14,
            16,
            18,
            20,
            24,
            30,
            36,
            42,
            48,
            60,
            72,
            84,
            96,
            120,
        )
        self.wsize.configure(values=sizes)

        families = sorted(tk.font.families())
        values = [""] + PREDEFINED_FONTS + families
        self.wfamily.configure(values=values)

    def _convert_to_form(self):
        if self._fp_cb is None:
            self._fp_cb = self.after_idle(self._convert_to_form_after)

    def _convert_to_form_after(self):
        self._form_clear()
        fdesc = tkfontstr_to_dict(self._font_var.get())
        key = "family"
        if fdesc[key]:
            self.family_var.set(fdesc[key])
        key = "size"
        if fdesc[key]:
            self.size_var.set(fdesc[key])
        modifiers = fdesc["modifiers"]
        if "bold" in modifiers:
            self.w_var.set(True)
        if "italic" in modifiers:
            self.s_var.set(True)
        if "underline" in modifiers:
            self.u_var.set(True)
        if "overstrike" in modifiers:
            self.o_var.set(True)
        self._form_process()

    def _value_from_form(self) -> str:
        """Process form inputs and return a tkfont description"""
        family = self.family_var.get()
        size = self.size_var.get()
        weight = self.w_var.get()
        slant = self.s_var.get()
        underline = self.u_var.get()
        overstrike = self.o_var.get()
        modifiers = []
        if weight:
            modifiers.append("bold")
        if slant:
            modifiers.append("italic")
        if underline:
            modifiers.append("underline")
        if overstrike:
            modifiers.append("overstrike")
        modifiers_str = " ".join(modifiers)
        font_tpl = "{{{family}}} {size} {{{modifiers}}}"
        return font_tpl.format(
            family=family, size=size, modifiers=modifiers_str
        )

    def _form_clear(self):
        self.family_var.set("")
        self.size_var.set("")
        self.w_var.set(False)
        self.s_var.set(False)
        self.u_var.set(False)
        self.o_var.set(False)

    def _form_process(self):
        """Process form inputs and generate the new tkfont description.
        If the value changed from original generate event."""

        family: str = self.family_var.get()
        family = family.strip()
        new_value = None
        curr_value = self._font_var.get()

        if family == "":
            # clear options
            self._form_clear()
            new_value = ""
        else:
            size = self.size_var.get()
            if size == "":
                self.size_var.set(self.FONT_SIZE_DEFAULT)
            new_value = self._value_from_form()
        if curr_value != new_value:
            self._font_var.set(new_value)
            self.event_generate(self.EVENT_FONT_CHANGED)
            print(new_value)
        self._fp_cb = None

    def on_keypress_after(self):
        self.call_process_form()

    def on_keypress(self, event=None):
        if self._kp_cb is not None:
            self.after_cancel(self._kp_cb)
        self._kp_cb = self.after(self.KEY_PRESS_MS, self.on_keypress_after)

    def on_validate_size(self, p_entry_value):
        """Allow only numbers above 0"""

        valid = False

        if p_entry_value == "":
            valid = True
        else:
            with suppress(ValueError):
                value = int(p_entry_value)
                valid = value > 0

        return valid

    def call_process_form(self, event=None):
        if self._fp_cb is None:
            self._fp_cb = self.after_idle(self._form_process)

    def on_modifier_clicked(self):
        self._form_process()


if __name__ == "__main__":
    root = tk.Tk()
    widget = FontInput(root)
    widget.configure(
        value="{Helvetica} 12 {bold}",
        # state="disabled"
    )
    widget.pack(expand=True, fill="both")
    root.mainloop()



# encoding: utf-8
import re

RE_FONT = re.compile(
    "(?P<family>\\{\\w+(\\w|\\s)*\\}|\\w+)\\s?(?P<size>-?\\d+)?\\s?(?P<modifiers>\\{\\w+(\\w|\\s)*\\}|\\w+)?"
)


def tkfontstr_to_dict(fontstr):
    """Parse a string with a tk font format like '{Helvetica} 12 {bold}'
    and return a dict with keys family, size and modifiers."""

    font_dict = {
        "family": None,
        "size": None,
        "modifiers": None,
    }
    s = RE_FONT.search(fontstr)
    if s:
        g = s.groupdict()
        font_dict["family"] = g["family"].replace("{", "").replace("}", "")
        font_dict["size"] = g["size"]
        modifiers = g["modifiers"]
        if modifiers is not None:
            modifiers = modifiers.replace("{", "").replace("}", "")
            font_dict["modifiers"] = modifiers
    else:
        font_dict["family"] = fontstr
    return font_dict


def tkfontstr_to_tuple(fontstr, default_size=12):
    """Convert string with a tk font format like '{Helvetica} 12 {bold}'
    to a python tuple ('Helvetica', 12, 'bold')."""

    fdict = tkfontstr_to_dict(fontstr)
    font_desc = [
        fdict["family"],
    ]
    # Add default size
    size = fdict["size"]
    size = size if size is not None else default_size
    font_desc.append(size)

    key = "modifiers"
    if fdict[key] is not None:
        font_desc.append(fdict[key])

    return tuple(font_desc)



import tkinter as tk
import customtkinter as ctk
from customtkinter.windows.widgets.core_widget_classes import CTkBaseClass
from pygubu.api.v1 import BuilderObject
from pygubu.plugins.pygubu.designer.basehelpers import (
    ToplevelPreviewBaseBO,
    ToplevelPreviewFactory,
    ToplevelPreviewMixin,
)
from ..widgets import CTkFrameBO, CTkSegmentedButtonBO
from ..tabview import CTkTabviewBO


#
# Preview class for CTkFrame
#
class CTkFrameForPreview(ctk.CTkFrame):
    def winfo_children(self):
        # CTkFrame has a hidden canvas inside. So, to make it clickable on preview
        # we need a hack.
        return super(tk.Frame, self).winfo_children()


class CTkFramePreviewBO(CTkFrameBO):
    class_ = CTkFrameForPreview


#
# Preview class for Tabview
#
class CTkTabviewForPreview(ctk.CTkTabview):
    #    def winfo_children(self):
    #        return super(tk.Frame, self).winfo_children()

    def bind(self, sequence=None, func=None, add=None):
        return super(tk.Frame, self).bind(sequence, func, add)


class CTkTabviewForPreviewBO(CTkTabviewBO):
    class_ = CTkTabviewForPreview


#
# Preview classes for ctk.CTKSegmentedbutton:
#


class CTkSegmentedButtonForPreview(ctk.CTkSegmentedButton):
    def bind(self, sequence=None, func=None, add=None):
        # FIXME: this is not working...
        #        I can't select a segmented button in preview ㅜㅜ
        for child_name in self.winfo_children():
            child = self.nametowidget(child_name)
            child.bind(sequence, func, True)


class CTkSegmentedButtonForPreviewBO(CTkSegmentedButtonBO):
    class_ = CTkSegmentedButtonForPreview


#
# Preview classes for CTKToplevel
#
class CTKToplevelPreviewMixin:
    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        # configure properties not supported by CtkFrame but yes by CtkToplevel
        props = ("borderwidth", "highlightbackground", "highlightthickness")
        for pname in props:
            if pname in kw:
                super(CTkBaseClass, self).configure(**{pname: kw.pop(pname)})
        return super().configure(cnf, **kw)


CTKToplevelPreview = ToplevelPreviewFactory(
    "CTKToplevelPreview",
    (CTKToplevelPreviewMixin, ToplevelPreviewMixin, CTkFrameForPreview, object),
    {},
)


class CTkToplevelPreviewBO(ToplevelPreviewBaseBO):
    class_ = CTKToplevelPreview
    ro_properties = ToplevelPreviewBaseBO.ro_properties + (
        "background",
        "fg_color",
    )

    def _process_property_value(self, pname, value):
        if pname in ("width", "height"):
            return int(value)
        return super()._process_property_value(pname, value)


#
# Preview classes for CTK
#
class CTKPreviewMixin:
    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        # configure properties not supported by CtkFrame but yes by Ctk
        props = ("padx", "pady", "relief", "takefocus")
        for pname in props:
            if pname in kw:
                super(CTkBaseClass, self).configure(**{pname: kw.pop(pname)})
        return super().configure(cnf, **kw)


CTKPreview = ToplevelPreviewFactory(
    "CTKPreview",
    (CTKPreviewMixin, ToplevelPreviewMixin, CTkFrameForPreview, object),
    {},
)


class CTkPreviewBO(CTkToplevelPreviewBO):
    class_ = CTKPreview
    properties = ToplevelPreviewBaseBO.properties + ("appearance_mode",)
    ro_properties = ToplevelPreviewBaseBO.ro_properties + ("fg_color",)

    def _set_property(self, target_widget, pname, value):
        if pname == "appearance_mode":
            ctk.set_appearance_mode(value)
        elif pname == "color_theme":
            ctk.set_default_color_theme(value)
        else:
            return super()._set_property(target_widget, pname, value)



from pygubu.api.v1 import register_widget
from tkcalendar import Calendar

from ..tkcalendar import _designer_tab_label, _plugin_uid
from .basecalendar import CalendarBaseBO


class CalendarBO(CalendarBaseBO):
    class_ = Calendar
    OPTIONS_STANDARD = ("class_", "cursor")
    OPTIONS_SPECIFIC = ("borderwidth", "state")
    properties = (
        OPTIONS_STANDARD + OPTIONS_SPECIFIC + CalendarBaseBO.OPTIONS_CUSTOM
    )
    ro_properties = CalendarBaseBO.ro_properties + ("class_",)
    virtual_events = ("<<CalendarSelected>>", "<<CalendarMonthChanged>>")


_builder_id = f"{_plugin_uid}.Calendar"

register_widget(
    _builder_id,
    CalendarBO,
    "Calendar",
    ("ttk", _designer_tab_label),
)



import tkinter as tk
from datetime import datetime
from pygubu.api.v1 import BuilderObject


class CalendarBaseBO(BuilderObject):
    OPTIONS_CUSTOM = (
        "year",
        "month",
        "day",
        "firstweekday",
        "mindate",
        "maxdate",
        "showweeknumbers",
        "showothermonthdays",
        "date_pattern",
        "selectmode",
        "textvariable",
        "background",
        "foreground",
        "disabledbackground",
        "disabledforeground",
        "bordercolor",
        "headersbackground",
        "headersforeground",
        "selectbackground",
        "selectforeground",
        "disabledselectbackground",
        "disabledselectforeground",
        "normalbackground",
        "normalforeground",
        "weekendbackground",
        "weekendforeground",
        "othermonthforeground",
        "othermonthbackground",
        "othermonthweforeground",
        "othermonthwebackground",
        "disableddaybackground",
        "disableddayforeground",
        "tooltipforeground",
        "tooltipbackground",
        "tooltipalpha",
        "tooltipdelay",
    )
    ro_properties = ("year", "month", "day")

    def _process_property_value(self, pname, value):
        if pname in ("year", "month", "day", "tooltipdelay"):
            return int(value)
        if pname in ("showweeknumbers", "showothermonthdays"):
            return tk.getboolean(value)
        if pname == "tooltipalpha":
            return float(value)
        if pname in ("mindate", "maxdate"):
            return datetime.strptime(value, "%Y-%m-%d")
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname in ("showweeknumbers", "showothermonthdays"):
            return self._process_property_value(pname, value)
        if pname in ("mindate", "maxdate"):
            return f"""Calendar.datetime.strptime("{value}", "%Y-%m-%d")"""
        return super()._code_process_property_value(targetid, pname, value)



#!/usr/bin/python3
import tkinter as tk
import tkinter.ttk as ttk


#
# Base class definition
#
class FontInputUI(ttk.Frame):
    def __init__(self, master=None, **kw):
        super().__init__(master, **kw)
        self.wfamily = ttk.Combobox(self, name="wfamily")
        self.family_var = tk.StringVar()
        self.wfamily.configure(textvariable=self.family_var)
        self.wfamily.pack(expand=False, fill="x", side="top")
        self.wfamily.bind(
            "<<ComboboxSelected>>", self.call_process_form, add=""
        )
        self.wfamily.bind("<FocusOut>", self.call_process_form, add="")
        self.wfamily.bind("<KeyPress>", self.on_keypress, add="")
        frame3 = ttk.Frame(self)
        frame3.configure(height=200, padding="0 1p 0 0", width=200)
        self.wsize = ttk.Spinbox(frame3, name="wsize")
        self.size_var = tk.StringVar()
        self.wsize.configure(
            from_=2, textvariable=self.size_var, to=900, validate="key", width=4
        )
        self.wsize.grid(column=0, row=0, sticky="ew")
        _validatecmd = (self.wsize.register(self.on_validate_size), "%P")
        self.wsize.configure(validatecommand=_validatecmd)
        self.wsize.bind("<<Decrement>>", self.call_process_form, add="")
        self.wsize.bind("<<Increment>>", self.call_process_form, add="")
        self.wsize.bind("<KeyPress>", self.on_keypress, add="")
        self.wweight = ttk.Checkbutton(frame3, name="wweight")
        self.w_var = tk.BooleanVar()
        self.wweight.configure(
            compound="center",
            style="weight.FontInput.Toolbutton",
            variable=self.w_var,
            width=-1,
        )
        self.wweight.grid(column=1, row=0)
        self.wweight.configure(command=self.on_modifier_clicked)
        self.wslant = ttk.Checkbutton(frame3, name="wslant")
        self.s_var = tk.BooleanVar()
        self.wslant.configure(
            compound="center",
            style="slant.FontInput.Toolbutton",
            variable=self.s_var,
            width=-1,
        )
        self.wslant.grid(column=2, row=0)
        self.wslant.configure(command=self.on_modifier_clicked)
        self.wunderline = ttk.Checkbutton(frame3, name="wunderline")
        self.u_var = tk.BooleanVar()
        self.wunderline.configure(
            compound="center",
            style="underline.FontInput.Toolbutton",
            variable=self.u_var,
            width=-1,
        )
        self.wunderline.grid(column=3, row=0)
        self.wunderline.configure(command=self.on_modifier_clicked)
        self.woverstrike = ttk.Checkbutton(frame3, name="woverstrike")
        self.o_var = tk.BooleanVar()
        self.woverstrike.configure(
            compound="center",
            style="overstrike.FontInput.Toolbutton",
            variable=self.o_var,
            width=-1,
        )
        self.woverstrike.grid(column=4, row=0)
        self.woverstrike.configure(command=self.on_modifier_clicked)
        frame3.pack(expand=False, fill="x", side="top")
        frame3.columnconfigure(0, weight=1)
        frame3.columnconfigure(1, pad="6p", uniform="b")
        frame3.columnconfigure(2, uniform="b")
        frame3.columnconfigure(3, uniform="b")
        frame3.columnconfigure(4, uniform="b")
        self.configure(height=25, width=100)
        self.pack(expand=True, fill="x", side="top")

    def call_process_form(self, event=None):
        pass

    def on_keypress(self, event=None):
        pass

    def on_validate_size(self, p_entry_value):
        pass

    def on_modifier_clicked(self):
        pass


if __name__ == "__main__":
    root = tk.Tk()
    widget = FontInputUI(root)
    widget.pack(expand=True, fill="both")
    root.mainloop()



from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from ..customtkinter import _designer_tab_label, _plugin_uid
from .ctkbase import CTkBaseMixin, GCONTAINER

from customtkinter import CTkScrollableFrame


class CTkScrollableFrameBO(CTkBaseMixin, BuilderObject):
    class_ = CTkScrollableFrame
    container = True
    # CTkScrollableFrame does some weird things
    # with layout so I disable container layout here on purpose.
    container_layout = False
    properties = (
        "width",
        "height",
        "corner_radius",
        "border_width",
        "bg_color",
        "fg_color",
        "border_color",
        "scrollbar_fg_color",
        "scrollbar_button_color",
        "scrollbar_button_hover_color:",
        "label_fg_color",
        "label_text_color",
        "label_text",
        "label_font",
        "label_anchor",
        "orientation",
    )
    ro_properties = ("orientation",)


_builder_uid = f"{_plugin_uid}.CTkScrollableFrame"
register_widget(
    _builder_uid,
    CTkScrollableFrameBO,
    "CTkScrollableFrame",
    ("ttk", _designer_tab_label),
    group=GCONTAINER,
)

register_custom_property(
    _builder_uid,
    "label_anchor",
    "choice",
    values=(
        "",
        "n",
        "ne",
        "nw",
        "e",
        "w",
        "s",
        "se",
        "sw",
        "center",
    ),
    state="readonly",
)

register_custom_property(
    _builder_uid,
    "orientation",
    "choice",
    values=("vertical", "horizontal"),
    default_value="vertical",
    state="readonly",
)



import tkinter as tk
import tkinter.ttk as ttk
import weakref
from abc import ABC, ABCMeta, abstractmethod
from collections import defaultdict
from contextlib import suppress
from dataclasses import dataclass, field
from typing import Optional, DefaultDict


class StyleDefinitionMeta(ABCMeta):
    managers = []
    instances = weakref.WeakKeyDictionary()

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if name != "IStyleManager":
            StyleDefinitionMeta.managers.append(cls)

    def __call__(cls, *args, **kwargs):
        inst = super().__call__(*args, **kwargs)
        StyleDefinitionMeta.instances[inst] = None

        return inst


@dataclass
class RootInfo:
    current_theme: Optional[str] = None
    configured_themes: DefaultDict[str, list] = field(
        default_factory=lambda: defaultdict(list)
    )  # theme > [definitions]


class StyleManager:
    CONFIGURED_THEMES = {}
    EVENT_THEME_CHANGED = "<<StyleManager:ThemeChanged>>"
    STYLE_CLASS = ttk.Style

    @classmethod
    def definition_exists(cls, definition_uid: str, root: tk.Tk):
        curr_theme = root.tk.eval("return $ttk::currentTheme")
        exists = False
        info: RootInfo = cls.CONFIGURED_THEMES.get(root, None)
        if info and definition_uid in info.configured_themes[curr_theme]:
            exists = True

        return exists

    @classmethod
    def register_root(cls, definition_uid: str, root: tk.Tk, curr_theme: str):
        if root not in cls.CONFIGURED_THEMES:
            info = RootInfo(curr_theme)
            info.configured_themes[curr_theme].append(definition_uid)
            cls.CONFIGURED_THEMES[root] = info

            root.bind_class(
                "TFrame",
                "<<ThemeChanged>>",
                lambda e, r=root: cls._theme_change_monitor(root),
                add=True,
            )

    @classmethod
    def _theme_change_monitor(cls, root: tk.Tk):
        info: RootInfo = cls.CONFIGURED_THEMES[root]
        curr_theme = root.tk.eval("return $ttk::currentTheme")
        # print(f"{curr_theme=} root {id(root)=}")
        if info.current_theme != curr_theme:
            info.current_theme = curr_theme
            cls.reconfigure_all(root, info)
            root.event_generate(cls.EVENT_THEME_CHANGED)
        return "break"

    @classmethod
    def reconfigure_all(cls, root: tk.Tk, info: RootInfo):
        """Call reconfigure for all definitions in root.
        The call is done only if definition was no applied
        previously in current theme.
        """
        definitions = info.configured_themes[info.current_theme]
        style = cls.STYLE_CLASS(root)
        for instance in StyleDefinitionMeta.instances:
            if instance.UID not in definitions:
                instance.reconfigure(style)


class IStyleDefinition(ABC, metaclass=StyleDefinitionMeta):
    """Helps to manage a custom style definition."""

    @property
    @abstractmethod
    def UID() -> str:
        """This property will be supplied by the inheriting
        classes individually.
        It is an unique identifier for the style or group of
        styles that will manage.
        """
        ...

    def __init__(self):
        self.style_managed_externally = False

    def is_managed_externally(self, master: tk.Widget):
        """Determine if style was defined externally.
        Default implementation returns False.

        If style is managed externally, this class just
        drop management of this definition.
        """
        return False

    def initialize(self, master: tk.Widget):
        root = master.winfo_toplevel()
        initialized = StyleManager.definition_exists(self.UID, root)

        if initialized or self.style_managed_externally:
            return

        if self.is_managed_externally(master):
            self.style_managed_externally = True
            return

        style = StyleManager.STYLE_CLASS(root)
        self.setup(style)

        StyleManager.register_root(self.UID, root, style.theme_use())

    def reconfigure(self, style: ttk.Style) -> None:
        """Reconfigure style in current theme.
        For simple styles just call style_setup again (default)
        For complex styles, redefine method in the subclass.
        """
        if self.style_managed_externally:
            return
        self.setup(style)

    @abstractmethod
    def setup(self, style: ttk.Style) -> None:
        """Setup your custom style here."""
        ...



from pygubu.i18n import _
from pygubu.api.v1 import register_custom_property
from pygubu.plugins.tkinterweb import _plugin_uid


_builder_all = f"{_plugin_uid}.*"
_htmlFrame = f"{_plugin_uid}.HtmlFrame"
_htmlLabel = f"{_plugin_uid}.HtmlLabel"
_notebook = f"{_plugin_uid}.Notebook"
_colourSelector = f"{_plugin_uid}.ColourSelector"


plugin_properties = {
    "colour": dict(
        buid=_colourSelector, editor="colorentry", default_value="white"
    ),
    "height": dict(
        buid=_notebook,
        editor="dimensionentry",
        default_value=200,
    ),
    "horizontal_scrollbar": dict(
        buid=_htmlFrame,
        editor="choice",
        values=("", "true", "false", "auto"),
        state="readonly",
    ),
    "messages_enabled": dict(
        buid=_builder_all,
        editor="choice",
        values=("", "false", "true"),
        state="readonly",
        help=_("Show debugging messages on console output."),
    ),
    "text": dict(
        buid=_htmlLabel,
        editor="text",
    ),
    "vertical_scrollbar": dict(
        buid=_htmlFrame,
        editor="choice",
        values=("", "true", "false", "auto"),
        state="readonly",
    ),
    "width": dict(
        buid=_notebook,
        editor="dimensionentry",
        default_value=200,
    ),
}

for prop in plugin_properties:
    definitions = plugin_properties[prop]
    if isinstance(definitions, dict):
        definitions = [definitions]
    for definition in definitions:
        builders = definition.pop("buid", _builder_all)
        if isinstance(builders, str):
            builders = [builders]
        editor = definition.pop("editor", "entry")
        for builder_uid in builders:
            register_custom_property(builder_uid, prop, editor, **definition)



import tkinter as tk
import tkinter.ttk as ttk
import logging

from pygubu.utils.widget import iter_to_toplevel
from pygubu.widgets.dockfw.slotframe import SlotIndicator


logger = logging.getLogger(__name__)


class IDockFrame:
    ...


class IDockPane:
    ...


class IDockWidget:
    ...


def tab_below_mouse(noteb: ttk.Notebook, rootx: int, rooty: int):
    # Is calculated this way because mouse motion events
    # return negative values with respect to where the drag started.
    x = rootx - noteb.winfo_rootx()
    y = rooty - noteb.winfo_rooty()
    tab_index = noteb.tk.call(noteb._w, "identify", "tab", x, y)
    if not isinstance(tab_index, int):
        tab_index = None
    return tab_index


class DockingFramework:
    curr_dock = None
    curr_dpane = None
    curr_dwidget = None
    bmouse_dock = None
    bmouse_dpane = None
    bmouse_dwidget = None
    source_dwidget = None
    source_tab_clicked = None
    moving = False
    indicator_active = None
    cursor_default = "arrow"
    cursor_moving = "hand1"
    cursor_tab_target = "sb_down_arrow"
    cursor_showing = None
    indicator_bg_color = "#989cec"

    @classmethod
    def init_binding(cls, widget: tk.Widget):
        widget.bind("<Button-1>", cls.drag_start, add=True)
        widget.bind("<B1-Motion>", cls.drag_motion, add=True)
        widget.bind("<ButtonRelease-1>", cls.drag_end, add=True)

    @classmethod
    def get_targets_below_mouse(cls, event: tk.Event):
        dock = None
        pane = None
        widget = None

        widget_below = event.widget.winfo_containing(event.x_root, event.y_root)

        for w in iter_to_toplevel(widget_below):
            if isinstance(w, IDockFrame) and dock is None:
                dock = w
                break
            if isinstance(w, IDockPane) and pane is None:
                pane = w
            if isinstance(w, IDockWidget) and widget is None:
                widget = w
            if isinstance(w, ttk.Notebook) and widget is None:
                selected = None
                if event.type == tk.EventType.ButtonPress:
                    tab_clicked = w.tk.call(
                        w._w, "identify", "tab", event.x, event.y
                    )
                    if isinstance(tab_clicked, int):
                        tab_id = w.tabs()[tab_clicked]
                        selected = w.nametowidget(tab_id)
                nb_tabs = w.tabs()
                if selected is None and nb_tabs:
                    tab_id = nb_tabs[0]
                    selected = w.nametowidget(tab_id)
                if isinstance(selected, IDockWidget):
                    widget = selected
        if dock and widget and pane is None:
            pane = widget.parent_pane
        cls.bmouse_dock = dock
        cls.bmouse_dpane = pane
        cls.bmouse_dwidget = widget

    @classmethod
    def drag_start(cls, event: tk.Event):
        # Only start movement if a tab is clicked
        cls.source_tab_clicked = tab_below_mouse(
            event.widget, event.x_root, event.y_root
        )
        if cls.source_tab_clicked is not None:
            cls.get_targets_below_mouse(event)
            cls.curr_dock = cls.bmouse_dock
            cls.curr_dpane = cls.bmouse_dpane
            cls.curr_dwidget = cls.bmouse_dwidget

    @classmethod
    def drag_motion(cls, event: tk.Event):
        if cls.curr_dock is None:
            # avoid errors on pygubu-designer preview
            return

        show_cursor = cls.cursor_moving
        cls._handle_cursor_moving()

        # If drag ends in a menu, a key error is produced.
        widget_below = None
        try:
            widget_below = event.widget.winfo_containing(
                event.x_root, event.y_root
            )
        except KeyError:
            pass
        if not widget_below:
            return

        cls.get_targets_below_mouse(event)
        if cls.bmouse_dpane is not None and cls.curr_dpane != cls.bmouse_dpane:
            if (
                cls.curr_dpane is not None
                and cls.curr_dpane != cls.curr_dock.main_pane
            ):
                cls.curr_dpane.indicators_visible(False)
            cls.curr_dpane = cls.bmouse_dpane
            cls.curr_dpane.indicators_visible(True)
        if (
            cls.bmouse_dwidget is not None
            and cls.curr_dwidget is not None
            and cls.curr_dwidget != cls.bmouse_dwidget
        ):
            cls.curr_dwidget.indicators_visible(False)
            cls.curr_dwidget = cls.bmouse_dwidget
            if cls.curr_dwidget != cls.source_dwidget:
                cls.curr_dwidget.indicators_visible(True)

        last_indicator = cls.indicator_active
        cls.indicator_active = None
        if isinstance(widget_below, SlotIndicator):
            cls.indicator_active = widget_below

        indicator = cls.indicator_active
        if indicator:
            if not hasattr(indicator, "_ocolor"):
                indicator._ocolor = indicator.cget("background")
            indicator.configure(background=cls.indicator_bg_color)
            show_cursor = indicator.cget("cursor")
        if (
            last_indicator
            and last_indicator != indicator
            and last_indicator.winfo_exists()
        ):
            last_indicator.configure(background=last_indicator._ocolor)

        show_cursor = cls._handle_nootbook(event, widget_below, show_cursor)

        if cls.cursor_showing != show_cursor:
            cls.curr_dock.configure(cursor=show_cursor)

    @classmethod
    def _handle_cursor_moving(cls):
        if cls.moving is False:
            cls.source_dwidget = cls.curr_dwidget
            cls.curr_dock.indicators_visible(True)
            cls.curr_dpane.indicators_visible(True)
            cls.moving = True

    @classmethod
    def _handle_nootbook(cls, event, widget_below, show_cursor):
        if cls.bmouse_dwidget and isinstance(widget_below, ttk.Notebook):
            tab_clicked = tab_below_mouse(
                widget_below, event.x_root, event.y_root
            )
            if cls.bmouse_dwidget == cls.source_dwidget:
                if (
                    tab_clicked is not None
                    and tab_clicked != cls.source_tab_clicked
                ):
                    show_cursor = show_cursor = cls.cursor_tab_target
            elif tab_clicked is not None:
                show_cursor = cls.cursor_tab_target
        return show_cursor

    @classmethod
    def drag_end(cls, event):
        if cls.curr_dock:
            if cls.moving:
                cls.curr_dock.indicators_visible(False)
                if cls.curr_dpane:
                    cls.curr_dpane.indicators_visible(False)
                if cls.curr_dwidget:
                    cls.curr_dwidget.indicators_visible(False)
                # Execute move if any
                if cls.source_dwidget is not None:
                    cls.execute_move(event)

            cls.curr_dock.configure(cursor=cls.cursor_default)
        cls.moving = False
        cls.curr_dwidget = None

    @classmethod
    def execute_move(cls, event: tk.Event):
        widget_below = None
        # If drag ends in a menu, a key error is produced.
        try:
            widget_below = event.widget.winfo_containing(
                event.x_root, event.y_root
            )
        except KeyError:
            pass
        dock = cls.source_dwidget.maindock
        relative_to = None
        side = None
        if cls.indicator_active:
            side = cls.indicator_active.side
            target = cls.indicator_active.nametowidget(
                cls.indicator_active.winfo_parent()
            )
            relative_to = "pane" if isinstance(target, IDockPane) else "widget"
            if relative_to == "pane":
                dock._move_into_pane_side(
                    cls.source_dwidget, cls.curr_dpane, side
                )
            else:
                dock._move_to_widget_side(
                    cls.source_dwidget, cls.curr_dwidget, side
                )
            return
        if cls.curr_dwidget and isinstance(widget_below, ttk.Notebook):
            move_inside = True
            tab_below = tab_below_mouse(
                widget_below, event.x_root, event.y_root
            )
            if tab_below is None:
                # No move
                logger.debug("No move posible.")
                return
            if cls.curr_dwidget == cls.source_dwidget:
                # Move in same pane
                if tab_below == cls.source_tab_clicked:
                    move_inside = False
            if move_inside:
                logger.debug(
                    "Move %s into group of %s, position %s",
                    cls.source_dwidget,
                    cls.curr_dwidget,
                    tab_below,
                )
                dock._move_into_widget_group(
                    cls.curr_dwidget, cls.source_dwidget, tab_below
                )
